package inf101.v18.connectfour.player;

import inf101.v18.connectfour.board.IBoard;
import inf101.v18.connectfour.board.IVerticalBoard;
import inf101.v18.connectfour.objects.Empty;
import inf101.v18.connectfour.objects.Token;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * An AI is a Player which must be asked for input directly.
 */
public class AI extends Player implements IAI {
    private final Random random = new Random();

    public AI(String name, Color color) {
        super(name, color);
    }

    @Override
    public int getColumn(IVerticalBoard b) {
        List<int[]> topPositions = getTopElements(b);
        IVerticalBoard copy = b.copyAll();
        //Checks if the AI can win
        int i = winRow(copy, topPositions, this)[0];
        if (i != -1) {
            return i;
        }
        IPlayer nextPlayer = copy.nextPlayer();
        //Checks if the AI can prevent a winning move
        while (nextPlayer != this) {
            i = winRow(copy, topPositions, nextPlayer)[0];
            if (i != -1) {
                return i;
            }
            nextPlayer = copy.nextPlayer();
        }
        do {
            i = random.nextInt(b.getWidth());
        } while (!b.isFull() && !(b.get(i, b.getHeight()-1) instanceof Empty));
        return i;
    }

    /**
     * Retrieves a list of all possible drop locations.
     *
     * @param b A board
     * @return  A list of positions
     */
    private List<int[]> getTopElements(IVerticalBoard b) {
        List<int[]> postitions = new ArrayList<>();
        for (int x = 0; x < b.getWidth(); x++) {
            int y = b.getTopObject(x);
            if (y < b.getHeight() - 1 && b.get(x, y + 1) instanceof Empty) {
                postitions.add(new int[]{x, y});
            }
        }
        return postitions;
    }

    /**
     * Retrieves a list of all empty locations.
     *
     * @param b A board
     * @return  A list of positions
     */
    private List<int[]> getEmpty(IBoard b) {
        List<int[]> postitions = new ArrayList<>();
        for (int i = 0; i < b.getWidth(); i++) {
            for (int j = 0; j < b.getHeight(); j++) {
                if (b.get(i, j) instanceof Empty) {
                    postitions.add(new int[]{i, j});
                }
            }
        }
        return postitions;
    }

    /**
     * Tries to simulate all possible moves for a player, until it finds a winning move.
     *
     * @param b         A board
     * @param positions A list of positions on the board
     * @param  player   The player to simulate
     * @return          A column index or -1 if the player can't win.
     */
    private int[] winRow(IBoard b, List<int[]> positions, IPlayer player) {
        for (int[] coords : positions) {
            int x = coords[0];
            int y = coords[1];
            if (b instanceof IVerticalBoard) {
                b.set(x, y + 1, new Token(player));
                if (b.won()) {
                    return new int[]{x};
                }
                b.set(x, y + 1, new Empty());
            } else {
                b.set(x, y, new Token(player));
                if (b.won()) {
                    return new int[]{x, y};
                }
                b.set(x, y, new Empty());
            }
        }
        return new int[]{-1};
    }

    @Override
    public int[] getXY(IBoard b) {
        List<int[]> emptyPositions = getEmpty(b);
        IBoard copy = b.copyAll();
        //Checks if the AI can win
        int[] coords = winRow(copy, emptyPositions, this);
        if (coords[0] != -1) {
            return coords;
        }
        IPlayer nextPlayer = copy.nextPlayer();
        //Checks if the AI can prevent a winning move
        while (nextPlayer != this) {
            coords = winRow(copy, emptyPositions, nextPlayer);
            if (coords[0] != -1) {
                return coords;
            }
            nextPlayer = copy.nextPlayer();
        }
        int x, y;
        do {
            x = random.nextInt(b.getWidth());
            y = random.nextInt(b.getHeight());
        } while (!b.isFull() && !(b.get(x, y) instanceof Empty));
        return new int[]{x, y};
    }
}
